//-----------------------------------------------------------------------
// File : AsShaderHelper.h
// Desc : Asura Shader Helper
// Copyright(c) Project Asura All right reserved.
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// History
//-----------------------------------------------------------------------
// <---> 2010/12/04 [Pocol] VK쐬
// <End>

#ifndef __ASURA_SHADER_HELPER_H__
#define __ASURA_SHADER_HELPER_H__

#define _GRAPHICS_API_OPENGL

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <Cg/Cg.h>


//-----------------------------------------------------------------------
// Linker
//-----------------------------------------------------------------------
#pragma comment( lib, "cg.lib" )


namespace Schaft {
namespace Asura {
namespace Graphics {

/////////////////////////////////////////////////////////////////////////
// ICgEffect class
/////////////////////////////////////////////////////////////////////////
class ICgEffect
{
	//-------------------------------------------------------------------
	// list of friend classes
	//-------------------------------------------------------------------
	/* NOTHING */

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
#if   defined(_GRAPHICS_API_OPENGL)
	virtual bool			Create( const char* code, const char **args ) = 0;
	virtual bool			CreateFromFile( const char* filename, const char **args ) = 0;
#elif defined(_GRAPHICS_API_DIRECTX11)
    virtual bool            Create( ID3D11Device* pDevice, const char* code, const char **args  = 0;
    virtual bool            CreateFromFile( ID3DX11Device* pDevice, const char* code, const char **args ) = 0;
#endif
	virtual void			Release() = 0;
	virtual unsigned int	GetNumPasses( unsigned int tech ) = 0;
	virtual unsigned int	GetNumTechniques() = 0;
	virtual CGpass			GetPass( unsigned int tech_index, unsigned int pass_index ) = 0;
	virtual CGpass			GetPass( unsigned int tech_index, const char* pass_name ) = 0;
    virtual const char*     GetPassName( unsigned int tech_index, unsigned int pass_index ) = 0;
	virtual CGtechnique		GetTechnique( unsigned int index ) = 0;
	virtual CGtechnique		GetTechnique( const char* name ) = 0;
	virtual unsigned int    GetTechniqueIndex( const char *name ) = 0;
    virtual const char*     GetTechniqueName( unsigned int index ) = 0;
	virtual CGparameter		GetParameter( const char* name ) = 0;

protected:
	//-------------------------------------------------------------------
	// protected variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	/* NOTHING */

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};

/////////////////////////////////////////////////////////////////////////
// CgEffectFactory class
/////////////////////////////////////////////////////////////////////////
class CgEffectFactory
{
	//-------------------------------------------------------------------
	// list of friend classes
	//-------------------------------------------------------------------
	/* NOTHING */

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
	static ICgEffect* Create();

protected:
	//-------------------------------------------------------------------
	// protected variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	/* NOTHING */

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};

} // namespace Graphics
} // namespace Asura
} // namespace Schaft

#endif // __ASURA_SHADER_HELPER_H__