//----------------------------------------------------------------------
// File : AsTexture.h
// Desc : Asura Texture
// Copyright(c) 2010 Project Asura All right reserved.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// History
//----------------------------------------------------------------------
// <---> 2010/11/28 [Pocol] VK쐬.
// <End>

#ifndef __ASURA_TEXTURE_H__
#define __ASURA_TEXTURE_H__

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include "ASDef.h"

namespace Schaft {
namespace Asura {
namespace Graphics {

/////////////////////////////////////////////////////////////////////////
// ITexture class
/////////////////////////////////////////////////////////////////////////
class ITexture
{
	//-------------------------------------------------------------------
	// list of friend classes
	//-------------------------------------------------------------------
	/* NOTHING */

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	enum
	{
		TEXTURE_FORMAT_RGB,
		TEXTURE_FORMAT_RGBA,
		TEXTURE_FORMAT_DXT1,
        TEXTURE_FORMAT_DXT2,
		TEXTURE_FORMAT_DXT3,
        TEXTURE_FORMAT_DXT4,
		TEXTURE_FORMAT_DXT5,
	};

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
	virtual void     Release() = 0;
	virtual uint32_t GetWidth() = 0;
	virtual uint32_t GetHeight() = 0;
	virtual uint16_t GetBitsPerPixel() = 0;
	virtual uint32_t GetFormat() = 0;
	virtual uint32_t GetNumMipmaps() = 0;
	virtual uint32_t GetSize() = 0;
	virtual char*    GetPixels() = 0;
	virtual char     GetPixel( uint32_t index ) = 0;

protected:
	//-------------------------------------------------------------------
	// protected variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	/* NOTHING */

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};

/////////////////////////////////////////////////////////////////////////
// ITextureLoader class
/////////////////////////////////////////////////////////////////////////
class ITextureLoader
{
	//-------------------------------------------------------------------
	// list of friend classes
	//-------------------------------------------------------------------
	/* NOTHING */

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
	virtual bool Load( const char* pFilename ) = 0;
	virtual bool IsLoading() = 0;
	virtual bool Parse() = 0;
	virtual ITexture* GetTexture() = 0;
	virtual void Release() = 0;

protected:
	//--------------------------------------------------------------------
	// protected variables
	//--------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	/* NOTHING */

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};

/////////////////////////////////////////////////////////////////////////
// TextureLoaderFactory class
/////////////////////////////////////////////////////////////////////////
class TextureLoaderFactory
{
	//-------------------------------------------------------------------
	// list of friend classes
	//-------------------------------------------------------------------
	/* NOTHING */

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	enum
	{
		TEXTURE_TYPE_BMP = 0,
		TEXTURE_TYPE_TGA,
		TEXTURE_TYPE_DDS,
		NUM_TEXTURE_TYPE,
	};

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
	static ITextureLoader* Create( uint32_t type );

protected:
	//-------------------------------------------------------------------
	// protected variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	/* NOTHING */

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};


} // namespace Graphics
} // namespace Asura
} // namespace Schaft

#endif //__ASURA_TEXTURE_H__