//-----------------------------------------------------------------------
// File : AsShaderHelper.h
// Desc : Asura Shader Helper
// Copyright(c) Project Asura All right reserved.
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// History
//-----------------------------------------------------------------------
// <---> 2010/12/04 [Pocol] VK쐬
// <End>


//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include "AsShaderHelper.h"
#include <vector>
#include <string>
#include <map>
#include <cassert>
#if   defined(_GRAPHICS_API_OPENGL)
#include <Cg/CgGL.h>
#elif defined(_GRAPHICS_API_DIRECX11)
#include <Cg/CgD3D11.h>
#endif

//-----------------------------------------------------------------------
// Linker
//-----------------------------------------------------------------------
#if   defined(_GRAPHICS_API_OPENGL)
#pragma comment( lib, "cgGL.lib" )
#elif defined(_GRAPHICS_API_DIRECTX11)
#pragma comment( lib, "cgD3D11.lib" )
#endif

//-----------------------------------------------------------------------
// Defines
//-----------------------------------------------------------------------
#ifndef DLOG
#if defined(DEBUG) || defined(_DEBUG)
#define DLOG(x, ...) std::printf( "[File: %s, Line: %d] "x"\n", __FILE__, __LINE__, ##__VA_ARGS__ );
#else
#define DLOG(x, ...)
#endif
#endif

#ifndef ELOG
#define ELOG(x, ...) std::printf( "[File: %s, Line: %d] "x"\n", __FILE__, __LINE__, ##__VA_ARGS__ );
#endif

//-----------------------------------------------------------------------
// Using Statements
//-----------------------------------------------------------------------
/* NOTHING */


//-----------------------------------------------------------------------
// Forward Declarations
//-----------------------------------------------------------------------
/* NOTHING */

//-----------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------
CGcontext gContext = 0;


//-----------------------------------------------------------------------
// Functions
//-----------------------------------------------------------------------

///----------------------------------------------------------------------
///<summary>
///CgG[`FbNR[obN֐
///</summary>
///----------------------------------------------------------------------
void CheckCgErrorCb()
{
	CGerror lastError = cgGetError();
	if ( lastError != CG_NO_ERROR )
	{
		ELOG( "CgError : error=%s, place=%s", cgGetErrorString( lastError ), cgGetLastListing(gContext) );
		assert( false );
	}
}

namespace Schaft {
namespace Asura {
namespace Graphics {

/////////////////////////////////////////////////////////////////////////
// CgTechnique class
/////////////////////////////////////////////////////////////////////////
class CgTechnique
{
	//-------------------------------------------------------------------
	// list of friend classes
	//-------------------------------------------------------------------
	/* NOTHING */

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
	CgTechnique();
	~CgTechnique();
	void			Initialize( CGtechnique tech );
	void			Release();
	unsigned int	GetNumPass();
	CGpass			GetPass( unsigned int index );
	CGpass			GetPass( const char *name );
    const char*     GetPassName( unsigned int index );
	CGtechnique		GetTechnique();
	const char*		GetTechniqueName();
	CgTechnique& operator = ( const CgTechnique &tech );

protected:
	//-------------------------------------------------------------------
	// protected variables
	//-------------------------------------------------------------------
	CGtechnique mTechnique;
	std::vector< CGpass > mPasses;

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	/* NOTHING */

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};


/////////////////////////////////////////////////////////////////////////
// CgEffect class
/////////////////////////////////////////////////////////////////////////
class CgEffect : public ICgEffect
{
	//-------------------------------------------------------------------
	// list of friend classes
	//-------------------------------------------------------------------
	/* NOTHING */

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
	CgEffect();
	virtual ~CgEffect();
#if   defined(_GRAPHICS_API_OPENGL)
	bool			Create( const char *code, const char **args );
	bool			CreateFromFile( const char *filename, const char **args );
#elif defined(_GRAPHICS_API_DIRECTX11)
    bool            Create( ID3D11Device* pDevice, const char* code, const char **args );
    bool            CreateFromFile( ID3D11Device* pDevice, const char* filename, const char **args );
#endif
	void			Release();
	unsigned int	GetNumPasses( unsigned int index );
	unsigned int	GetNumTechniques();
	CGpass			GetPass( unsigned int tech_index, unsigned int pass_index );
	CGpass			GetPass( unsigned int tech_index, const char *pass_name );
    const char*     GetPassName( unsigned int tech_index, unsigned int pass_index );
	CGtechnique		GetTechnique( unsigned int index );
	CGtechnique		GetTechnique( const char *name );
    const char*     GetTechniqueName( unsigned int index );
	unsigned int	GetTechniqueIndex( const char* name );
	CGparameter		GetParameter( const char *name );
	bool            IsCgError();

protected:
	//-------------------------------------------------------------------
	// protected variables
	//-------------------------------------------------------------------
	CGeffect					mEffect;
	CGcontext					mContext;
	std::vector< CgTechnique >	mTechniques;
	std::map< std::string, unsigned int > mTechMaps;

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	bool SetTechniques();

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};


//-----------------------------------------------------------------------
// CgTechnique Implementation
//-----------------------------------------------------------------------

///----------------------------------------------------------------------
///<summary>
///RXgN^
///</summary>
///----------------------------------------------------------------------
CgTechnique::CgTechnique()
{
	mTechnique = 0;
	mPasses.clear();
}

///----------------------------------------------------------------------
///<summary>
///fXgN^
///</summary>
///----------------------------------------------------------------------
CgTechnique::~CgTechnique()
{
	Release();
}

///----------------------------------------------------------------------
///<summary>
///s
///</summary>
///<param name="tech">ݒ肷eNjbN</param>
///----------------------------------------------------------------------
void CgTechnique::Initialize( CGtechnique tech )
{
	mTechnique = tech;

	//@pXݒ肷
	CGpass pass = cgGetFirstPass( mTechnique );
	for( ;; )
	{
		//@NULLɂȂI
		if ( pass == 0 )
			break;

		//@LȂpXȂΊi[
		if ( cgIsPass( pass ) == CG_TRUE )
		{
			mPasses.push_back( pass );
		}
		else
		{
			ELOG( "CgError : Invalid Pass. Skipping - pass=%s", cgGetPassName( pass ) );
		}
		pass = cgGetNextPass( pass );
	}
}

///----------------------------------------------------------------------
///<summary>
///
///</summary>
///----------------------------------------------------------------------
void CgTechnique::Release()
{
	mTechnique = 0;
	mPasses.clear();
}

///----------------------------------------------------------------------
///<summary>
///pX擾
///</summary>
///<returns>pXԋp</returns>
///----------------------------------------------------------------------
unsigned int CgTechnique::GetNumPass()
{
	return static_cast< unsigned int >( mPasses.size() );
}

///----------------------------------------------------------------------
///<summary>
///pX擾
///</summary>
///<param name="index">pX̃CfbNX</param>
///<returns>w肳ꂽCfbNX̃pXԋp</returns>
///----------------------------------------------------------------------
CGpass CgTechnique::GetPass( unsigned int index )
{
	return mPasses[index];
}

///----------------------------------------------------------------------
///<summary>
///pX擾
///</summary>
///<param name="name">pX</param>
///<returns>w肳ꂽOpXԋp</returns>
///----------------------------------------------------------------------
CGpass CgTechnique::GetPass( const char* name )
{
	return cgGetNamedPass( mTechnique, name );
}

///----------------------------------------------------------------------
///<summary>
///pX擾
///</summary>
///<param name="index">pX̃CfbNX</param>
///<returns>w肳ꂽpXԋp</returns>
///----------------------------------------------------------------------
const char* CgTechnique::GetPassName( unsigned int index )
{
    return cgGetPassName( mPasses[index] );
}

///----------------------------------------------------------------------
///<summary>
///eNjbN擾
///</summary>
///----------------------------------------------------------------------
CGtechnique CgTechnique::GetTechnique()
{
	return mTechnique;
}

///----------------------------------------------------------------------
///<summary>
///eNjbN擾
///</summary>
///<returns>eNjbNԋp</returns>
///----------------------------------------------------------------------
const char* CgTechnique::GetTechniqueName()
{
	return cgGetTechniqueName( mTechnique );
}

///----------------------------------------------------------------------
///<summary>
///Zq
///</summary>
///----------------------------------------------------------------------
CgTechnique& CgTechnique::operator = ( const CgTechnique &tech )
{
	mTechnique = tech.mTechnique;
	size_t size = tech.mPasses.size();
	mPasses.resize( size );
	for( size_t i=0; i<size; i++ )
	{
		mPasses[i] = tech.mPasses[i];
	}
	return (*this);
}

//-----------------------------------------------------------------------
// CgEffect Implementation
//-----------------------------------------------------------------------

///----------------------------------------------------------------------
///<summary>
///RXgN^
///</summary>
///----------------------------------------------------------------------
CgEffect::CgEffect()
{
	mEffect = 0;
	mContext = 0;
	mTechniques.clear();
}

///----------------------------------------------------------------------
///<summary>
///fXgN^
///</summary>
///----------------------------------------------------------------------
CgEffect::~CgEffect()
{
	Release();
}

#if defined(_GRAPHICS_API_OPENGL)
///----------------------------------------------------------------------
///<summary>
///񂩂GtFNg쐬
///</summary>
///<param name="code">GtFNgi[Ă镶</param>
///<param name="args">CgRpC֓n</param>
///<returns>쐬ɐꍇtrueԋp</returns>
///----------------------------------------------------------------------
bool CgEffect::Create( const char *code, const char **args )
{
	if ( mContext == 0 )
	{
		mContext = cgCreateContext();
        gContext = mContext;
		cgGLRegisterStates( mContext );
		if ( IsCgError() ) return false;
	}

#if defined(DEBUG) || defined(_DEBUG)
	cgSetErrorCallback( CheckCgErrorCb );
#endif

	mEffect = cgCreateEffect( mContext, code, args );
	if ( cgIsEffect( mEffect ) == CG_FALSE )
	{
		ELOG( "CgError : Invalid Effect." );
		return false;
	}
	if ( IsCgError() )
	{
		return false;
	}
	return SetTechniques();
}

///----------------------------------------------------------------------
///<summary>
///t@CGtFNg쐬
///</summary>
///<param name="filename">GtFNgt@C</param>
///<param name="args">CgRpC֓n</param>
///<returns>쐬ɐꍇtrueԋp</returns>
///----------------------------------------------------------------------
bool CgEffect::CreateFromFile( const char* filename, const char **args )
{
    if ( !mContext )
	{
		mContext = cgCreateContext();
        gContext = mContext;
		cgGLRegisterStates( mContext );
		if ( IsCgError() ) return false;
	}

#if defined(DEBUG) || defined(_DEBUG)
	cgSetErrorCallback( CheckCgErrorCb );
#endif

	mEffect = cgCreateEffectFromFile( mContext, filename, args );
    if ( IsCgError() ) 
    {
        return false;
    }
	else if ( cgIsEffect( mEffect ) == CG_FALSE )
	{
		ELOG( "CgError : Invalid Effect. filename=%s", filename );
		return false;
	}
	if ( IsCgError() )
	{
		return false;
	}
	return SetTechniques();
}
#elif defined(_GRAPHICS_API_DIRECTX11)
///----------------------------------------------------------------------
///<summary>
///񂩂GtFNg쐬
///</summary>
///<param name="pDevice">foCXC^tF[X</param>
///<param name="code">GtFNgi[Ă镶</param>
///<param name="args">CgRpC֓n</param>
///<returns>쐬ɐꍇtrueԋp</returns>
///----------------------------------------------------------------------
bool CgEffect::Create( ID3D11Device *pDevice, const char *code, const char **args )
{
    /*  */
    return false;
}

///----------------------------------------------------------------------
///<summary>
///t@CGtFNg쐬
///</summary>
///<param name="filename">GtFNgt@C</param>
///<param name="args">CgRpC֓n</param>
///<returns>쐬ɐꍇtrueԋp</returns>
///----------------------------------------------------------------------
bool CgEffect::CreateFromFile( ID3D11Device* pDevice, const char *code, const char **args )
{
    /*  */
    return false;
}
#endif

///----------------------------------------------------------------------
///<summary>
///eNjbNݒ肷
///</summary>
///<returns>ݒɎsꍇfalseԋp</returns>
///----------------------------------------------------------------------
bool CgEffect::SetTechniques()
{
	mTechMaps.clear();
	mTechniques.clear();
	CGtechnique tech = cgGetFirstTechnique( mEffect );
	
	for( ;; )
	{
		if ( cgValidateTechnique( tech ) == CG_FALSE )
		{
			ELOG( "CgError : Invalid Technique. Skipping - technique=%s", cgGetTechniqueName( tech ) );
		}
		else
		{
			std::string name = std::string( cgGetTechniqueName( tech ) );
			DLOG( "Info : Valid Technique. - name=%s, index=%d", name.c_str(), mTechniques.size() );
			CgTechnique cg_tech;
			cg_tech.Initialize( tech );
			mTechMaps.insert( std::pair< std::string, unsigned int >( name, mTechniques.size() ) );
			mTechniques.push_back( cg_tech );
		}

		tech = cgGetNextTechnique( tech );
		assert( !IsCgError() );

		if ( tech == 0 )
		{
			break;
		}
	}

	if ( mTechniques.size() == 0 )
	{
		ELOG( "CgError : No Valid Techniques." );
		return false;
	}
	if ( IsCgError() )
	{
		return false;
	}

	return true;
}

///----------------------------------------------------------------------
///<summary>
///
///</summary>
///----------------------------------------------------------------------
void CgEffect::Release()
{
	if ( mEffect )
	{
		cgDestroyEffect( mEffect );
		mEffect = 0;
	}
	if ( mContext )
	{
		cgDestroyContext( mContext );
		mContext = 0;
	}

	mTechMaps.clear();

	for( size_t i=0; i<mTechniques.size(); i++ )
	{
		mTechniques[i].Release();
	}
	mTechniques.clear();
}

///----------------------------------------------------------------------
///<summary>
///w肳ꂽeNjbÑCfbNX̃pX擾
///</summary>
///<param name="index">eNjbÑCfbNX</param>
///<returns>w肳ꂽeNjbÑpXԋp</returns>
///----------------------------------------------------------------------
unsigned int CgEffect::GetNumPasses( unsigned int index )
{
	return mTechniques[index].GetNumPass();
}

///----------------------------------------------------------------------
///<summary>
///eNjbN擾
///</summary>
///<returns>eNjbNԋp</returns>
///----------------------------------------------------------------------
unsigned int CgEffect::GetNumTechniques()
{
	return static_cast< unsigned int >( mTechniques.size() );
}

///----------------------------------------------------------------------
///<summary>
///pX擾
///</summary>
///<param name="tech_index">eNjbÑCfbNX</param>
///<param name="pass_index">pX̃CfbNX</param>
///<returns>w肳ꂽCfbNX̃pXԋp</returns>
///----------------------------------------------------------------------
CGpass CgEffect::GetPass( unsigned int tech_index, unsigned int pass_index )
{
	return mTechniques[tech_index].GetPass( pass_index );
}

///----------------------------------------------------------------------
///<summary>
///pX擾
///</summary>
///<param name="tech_index">eNjbÑCfbNX</param>
///<param name="pass_name">pX̖O</param>
///<returns>w肳ꂽpX̃pXԋp</returns>
///----------------------------------------------------------------------
CGpass CgEffect::GetPass( unsigned int tech_index, const char *pass_name )
{
	return mTechniques[tech_index].GetPass( pass_name );
}

///----------------------------------------------------------------------
///<summary>
///pX擾
///</summary>
///<param name="tech_index">eNjbÑCfbNX</param>
///<param name="pass_index">pX̃CfbNX</param>
///<returns>w肳ꂽpXԋp</returns>
///----------------------------------------------------------------------
const char* CgEffect::GetPassName( unsigned int tech_index, unsigned int pass_index )
{
    return mTechniques[tech_index].GetPassName( pass_index );
}

///----------------------------------------------------------------------
///<summary>
///eNjbN擾
///</summary>
///<param name="index">eNjbÑCfbNX</param>
///<returns>w肳ꂽCfbNX̃eNjbNԋp</returns>
///----------------------------------------------------------------------
CGtechnique CgEffect::GetTechnique( unsigned int index )
{
	return mTechniques[index].GetTechnique();
}

///----------------------------------------------------------------------
///<summary>
///eNjbN擾
///</summary>
///<param name="name">eNjbN</param>
///<returns>w肳ꂽeNjbÑeNjbNԋp</returns>
///----------------------------------------------------------------------
CGtechnique CgEffect::GetTechnique( const char *name )
{
	return cgGetNamedTechnique( mEffect, name );
}

///----------------------------------------------------------------------
///<summary>
///eNjbN擾
///</summary>
///<param name="index">eNjbÑCfbNX</param>
///<returns>w肳ꂽeNjbNԋp</returns>
///----------------------------------------------------------------------
const char* CgEffect::GetTechniqueName( unsigned int index )
{
    return mTechniques[index].GetTechniqueName();
}

///----------------------------------------------------------------------
///<summary>
///eNjbÑCfbNX擾
///</summary>
///<param name="name">eNjbN</param>
///<returns>
///w肳ꂽeNjbÑCfbNXԋp.
///Ȃꍇ̓I[f(-1)ԋp.
///</returns>
///----------------------------------------------------------------------
unsigned int CgEffect::GetTechniqueIndex( const char *name )
{
	std::map< std::string, unsigned int >::iterator itr;
	itr = mTechMaps.find( std::string( name ) );
	if ( itr != mTechMaps.end() )
	{
		return itr->second;
	}
	return -1;
}

///----------------------------------------------------------------------
///<summary>
///p[^擾
///</summary>
///<param name="name">p[^</param>
///<returns>w肳ꂽp[^̃p[^ԋp</returns>
///----------------------------------------------------------------------
CGparameter CgEffect::GetParameter( const char *name )
{
	return cgGetNamedEffectParameter( mEffect, name );
}

///----------------------------------------------------------------------
///<summary>
///CgG[𔻒肷
///</summary>
///<returns>CgG[ꍇtrueԋp</returns>
///----------------------------------------------------------------------
bool CgEffect::IsCgError()
{
	CGerror lastError = cgGetError();
	if ( lastError != CG_NO_ERROR )
	{
		ELOG( "CgError : error=%s, place=%s", cgGetErrorString( lastError ), cgGetLastListing( mContext ) );
		return true;
	}
	return false;
}

/////////////////////////////////////////////////////////////////////////
// CgEffectFactory class
/////////////////////////////////////////////////////////////////////////

///---------------------------------------------------------------------
///<summary>
///CgGtFNg𐶐
///</summary>
///---------------------------------------------------------------------
ICgEffect* CgEffectFactory::Create()
{
    return new CgEffect();
}



} // namespace Schaft
} // namespace Asura
} // namespace Graphics