//----------------------------------------------------------------------
// File : AsTextureLoaderDDS.h
// Desc : Asura DDS Texture Loader
// Copyright(c) 2010 Project Asura All right reserved.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// History
//----------------------------------------------------------------------
// <---> 2010/11/28 [Pocol] VK쐬.
// <End>

//----------------------------------------------------------------------
// Includes 
//----------------------------------------------------------------------
#include "AsTextureLoaderDDS.h"


//----------------------------------------------------------------------
// Defines
//----------------------------------------------------------------------
#ifndef DEBUG_LOG
#if 0
#define DEBUG_LOG DLOG
#else
#define DEBUG_LOG
#endif
#endif


namespace Schaft {
namespace Asura {
namespace Graphics {


//-----------------------------------------------------------------------
// Constant Variables
//-----------------------------------------------------------------------
const int D3D_FORMAT_DXT1		= '1TXD';
const int D3D_FORMAT_DXT2		= '2TXD';
const int D3D_FORMAT_DXT3		= '3TXD';
const int D3D_FORMAT_DXT4		= '4TXD';
const int D3D_FORMAT_DXT5		= '5TXD';
const int D3D_FORMAT_R8G8B8		= 20;
const int D3D_FORMAT_A8R8G8B8	= 21;
const int D3D_FORMAT_X8R8G8B8	= 22;
const int D3D_FORMAT_A8B8G8R8	= 32;
const int D3D_FORMAT_X8B8G8R8	= 33;
const long DDS_FOURCC			= 0x00000004;
const long DDS_RGB				= 0x00000040;
const long DDS_RGBA				= 0x00000041;


////////////////////////////////////////////////////////////////////////
// TextureLoaderDDS class
////////////////////////////////////////////////////////////////////////

///---------------------------------------------------------------------
///<summary>
///RXgN^
///</summary>
///---------------------------------------------------------------------
TextureLoaderDDS::TextureLoaderDDS() 
: TextureLoaderBase()
{
	/* DO_NOTHING */
}

///---------------------------------------------------------------------
///<summary>
///fXgN^
///</summary>
///---------------------------------------------------------------------
TextureLoaderDDS::~TextureLoaderDDS()
{
	Release();
}

///---------------------------------------------------------------------
///<summary>
///}WbN`FbN
///</summary>
///<param name="pFilename">`FbNt@C</param>
///<returns>DDSt@CȂtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::CheckMagic( const char * pFilename )
{
    FILE *pFile = fopen( pFilename, "rb" );
    if ( !pFile )
    {
        ELOG( "Error: Couldn't open %s.", pFilename );
        return false;
    }

    char magic[4];
    fread( &magic, sizeof(magic), 1, pFile );
    fclose( pFile );

    if ( strncmp( magic, "DDS ", 4 ) != 0 )
    {
        ELOG( "Error : %s is not a valid DDS file.", pFilename );
        return false;
    }

    return true;
}

///---------------------------------------------------------------------
///<summary>
///DDSt@C[h
///</summary>
///<param name="pFilename">[ht@C</param>
///<returns>[hɐtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::Load( const char *pFilename )
{
	//@}WbN`FbN
    if ( !CheckMagic( pFilename ) )
    {
		DLOG( "Info : CheckMagic() failed." );
        return false;
    }

	//@t@CJ
    mpFile = fopen( pFilename, "rb" );
    if ( !mpFile )
    {
        ELOG( "Error: Couldn't open %s.", pFilename );
        return false;
    }

	//@wb_ǂݎ
    char magic[4];
    fread( &magic, sizeof(magic), 1, mpFile );
    fread( &mDesc, sizeof(mDesc), 1, mpFile );

	//@ƍC~bv}bvݒ
    mTexture.SetWidth( mDesc.width );
    mTexture.SetHeight( mDesc.height );
	mTexture.SetNumMipmaps( mDesc.mipMapLevels );

	//@obt@TCY𒲂ׂ
    long cur = ftell( mpFile );
    fseek( mpFile, 0, SEEK_END );
    long end = ftell( mpFile );
    fseek( mpFile, cur, SEEK_SET );

	//@obt@TCYݒ
    mBufSize = end - cur;

	//@m
    try
    {
        mpBuf = new char [mBufSize];
    }
    catch( std::bad_alloc ba )
    {
        ELOG( "Error: Memory allocate failed." );
        fclose( mpFile );
        return false;
    }

	//@ǂݍ݃TCY
    mCurSize = 0;
    mRestSize = mBufSize;

    return true;
}

///---------------------------------------------------------------------
///<summary>
///rbg}XN𒲂ׂ
///</summary>
///<param name="r">R</param>
///<param name="g">G</param>
///<param name="b">B</param>
///<param name="a">A</param>
///<returns>
///w肳ꂽ}XNƓǂݍ݂}XNvĂtrueԋp
///</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::IsBitMask( long r, long g, long b, long a )
{
    return ( mDesc.format.redMask == r ) 
        && ( mDesc.format.greenMask == g )
        && ( mDesc.format.blueMask == b )
        && ( mDesc.format.alphaMask == a );
}

///---------------------------------------------------------------------
///<summary>
///64BittH[}bg͂
///</summary>
///<returns>͂ɐtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::Parse64Bit()
{
	bool result = false;

	/*  */

	return result;
}

///---------------------------------------------------------------------
///<summary>
///32BittH[}bg͂
///</summary>
///<returns>͂ɐtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::Parse32Bit()
{
	bool result = false;

    // A8 R8 G8 B8
    if ( IsBitMask( 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000 ) )
    {
        mTexture.SetFormat( ITexture::TEXTURE_FORMAT_RGBA );
        mTexture.SetBitsPerPixel( 8 );
        mTexture.Resize( mBufSize );

        for( uint32_t i=0; i<mBufSize; i+= 4 )
        {
            char a = mpBuf[i + 0];
            char r = mpBuf[i + 1];
            char g = mpBuf[i + 2];
            char b = mpBuf[i + 3];
            mTexture.SetPixel( i + 0, r );
            mTexture.SetPixel( i + 1, g );
            mTexture.SetPixel( i + 2, b );
            mTexture.SetPixel( i + 3, a );
        }
        SAFE_DELETE_ARRAY( mpBuf );
		result = true;
    }
    // X8 R8 G8 B8
    else if ( IsBitMask( 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000 ) )
    {
        mTexture.SetFormat( ITexture::TEXTURE_FORMAT_RGB );
        mTexture.SetBitsPerPixel( 8 );
        mTexture.Resize( mBufSize );

        for( uint32_t i=0; i<mBufSize; i+=3 )
        {
            char r = mpBuf[i + 0];
            char g = mpBuf[i + 1];
            char b = mpBuf[i + 2];
            mTexture.SetPixel( i + 0, r );
            mTexture.SetPixel( i + 1, g );
            mTexture.SetPixel( i + 2, b );
        }
        SAFE_DELETE_ARRAY( mpBuf );
		result = true;
    }
    // A8 B8 G8 R8
    else if ( IsBitMask( 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000 ) )
    {
        mTexture.SetFormat( ITexture::TEXTURE_FORMAT_RGBA );
        mTexture.SetBitsPerPixel( 8 );
        mTexture.Resize( mBufSize );

        for( uint32_t i=0; i<mBufSize; i+=4 )
        {
            char a = mpBuf[i + 0];
            char b = mpBuf[i + 1];
            char g = mpBuf[i + 2];
            char r = mpBuf[i + 3];
            mTexture.SetPixel( i + 0, r );
            mTexture.SetPixel( i + 1, g );
            mTexture.SetPixel( i + 2, b );
            mTexture.SetPixel( i + 3, a );
        }
        SAFE_DELETE_ARRAY( mpBuf );
		result = true;
    }
    // X8 B8 G8 R8
    else if ( IsBitMask( 0x000000ff, 0x0000ff00, 0x00ff0000, 0x00000000 ) )
    {
        mTexture.SetFormat( ITexture::TEXTURE_FORMAT_RGB );
        mTexture.SetBitsPerPixel( 8 );
        mTexture.Resize( mBufSize );

        for( uint32_t i=0; i<mBufSize; i+= 3 )
        {
            char b = mpBuf[i + 0];
            char g = mpBuf[i + 1];
            char r = mpBuf[i + 2];
            mTexture.SetPixel( i + 0, r );
            mTexture.SetPixel( i + 1, g );
            mTexture.SetPixel( i + 2, b );
        }
        SAFE_DELETE_ARRAY( mpBuf );
		result = true;
    }

	return result;
}

///---------------------------------------------------------------------
///<summary>
///24BittH[}bg͂
///</summary>
///<returns>͂ɐtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::Parse24Bit()
{
	bool result = false;

    // R8 G8 B8
    if ( IsBitMask( 0x00ff0000, 0x0000ff00, 0x000000ff, 0x00000000 ) )
    {
        mTexture.SetBitsPerPixel( 8 );
        mTexture.SetFormat( ITexture::TEXTURE_FORMAT_RGB );
        mTexture.Resize( mBufSize );

        for( uint32_t i=0; i<mBufSize; i+= 3 )
        {
            char r = mpBuf[i + 0];
            char g = mpBuf[i + 1];
            char b = mpBuf[i + 2];
            mTexture.SetPixel( i + 0, r );
            mTexture.SetPixel( i + 1, g );
            mTexture.SetPixel( i + 2, b );
        }
        SAFE_DELETE_ARRAY( mpBuf );
		result = true;
    }

	return result;
}

///---------------------------------------------------------------------
///<summary>
///16BittH[}bg͂
///</summary>
///<returns>͂ɐtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::Parse16Bit()
{
	bool result = false;

	/*  */

	return result;
}

///---------------------------------------------------------------------
///<summary>
///8BittH[}bg͂
///</summary>
///<returns>͂ɐtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::Parse8Bit()
{
	bool result = false;

	/*  */

	return result;
}

///---------------------------------------------------------------------
///<summary>
///DXTktH[}bg͂.
///\tgEFAfR[h͂Ȃ̂ŗv.
///</summary>
///<returns>͂ɐtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::ParseDXT()
{
	bool result = false;

    // DXT1
    if ( mDesc.format.fourCC == D3D_FORMAT_DXT1 )
    {
		mTexture.SetBitsPerPixel( 0 ); 
		mTexture.SetFormat( ITexture::TEXTURE_FORMAT_DXT1 );
		result = true;
    }
    // DXT2
    else if ( mDesc.format.fourCC == D3D_FORMAT_DXT2 )
    {
		mTexture.SetBitsPerPixel( 0 );
		mTexture.SetFormat( ITexture::TEXTURE_FORMAT_DXT2 );
		result = true;
    }
    // DXT3
    else if ( mDesc.format.fourCC == D3D_FORMAT_DXT3 )
    {
		mTexture.SetBitsPerPixel( 0 );
		mTexture.SetFormat( ITexture::TEXTURE_FORMAT_DXT3 );
		result = true;
    }
    // DXT4
    else if ( mDesc.format.fourCC == D3D_FORMAT_DXT4 )
    {
		mTexture.SetBitsPerPixel( 0 );
		mTexture.SetFormat( ITexture::TEXTURE_FORMAT_DXT4 );
		result = true;
    }
    // DXT5
    else if ( mDesc.format.fourCC == D3D_FORMAT_DXT5 )
    {
		mTexture.SetBitsPerPixel( 0 );
		mTexture.SetFormat( ITexture::TEXTURE_FORMAT_DXT5 );
		result = true;
    }

	if ( result )
	{
		mTexture.Resize( mBufSize );
		mTexture.SetPixels( mpBuf );
		SAFE_DELETE_ARRAY( mpBuf );
	}

	return result;
}

///---------------------------------------------------------------------
///<summary>
///tH[}bg̉͂
///</summary>
///<returns>YtH[}bgꍇtrueԋp</returns>
///---------------------------------------------------------------------
bool TextureLoaderDDS::Parse()
{
    bool result = false;

	if ( ParseDXT() )
	{
		DEBUG_LOG( "Info : ParseDXT() success." );
		result = true;
	}
	else if ( Parse64Bit() )
	{
		DEBUG_LOG( "Info : Parse64Bit() success." );
		result = true;
	}
	else if ( Parse32Bit() )
	{
		DEBUG_LOG( "Info : Parse32Bit() success." );
		result = true;
	}
	else if ( Parse24Bit() )
	{
		DEBUG_LOG( "Info : Parse24Bit() success." );
		result = true;
	}
	else if ( Parse16Bit() )
	{
		DEBUG_LOG( "Info : Parse16Bit() success." );
		result = true;
	}
	else if ( Parse8Bit() )
	{
		DEBUG_LOG( "Info : Parse8Bit() success." );
		result = true;
	}
	else
	{
		ELOG( "Error : Parse() failed." );
		SAFE_DELETE_ARRAY( mpBuf );
		mTexture.Release();
	}
    
    return result;
}


} // namespace Graphics
} // namespace Asura
} // namespace Schaft