//----------------------------------------------------------------------
// File : ASTextureImpl.h
// Desc : Asura Texture Implement
// Copyright(c) 2010 Project Asura All right reserved.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// History
//----------------------------------------------------------------------
// <---> 2010/11/28 [Pocol] VK쐬.
// <End>


#ifndef __ASURA_TEXTURE_LOADER_DDS_H__
#define __ASURA_TEXTURE_LOADER_DDS_H__

//----------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------
#include "AsTextureImpl.h"

namespace Schaft {
namespace Asura {
namespace Graphics {


//-----------------------------------------------------------------------
// Type definition
//-----------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////
// DDS_PIXELFORMAT structuer
/////////////////////////////////////////////////////////////////////////
typedef struct _DDS_PIXELFORMAT
{
	unsigned int size;
	unsigned int flags;
	unsigned int fourCC;
	unsigned int bpp;
	unsigned int redMask;
	unsigned int greenMask;
	unsigned int blueMask;
	unsigned int alphaMask;
} DDS_PIXELFORMAT;

/////////////////////////////////////////////////////////////////////////
// DDS_CAPS structure
/////////////////////////////////////////////////////////////////////////
typedef struct _DDS_CAPS
{
	unsigned int caps1;
	unsigned int caps2;
	unsigned int caps3;
	unsigned int caps4;
} DDS_CAPS;

/////////////////////////////////////////////////////////////////////////
// DDS_COLORKEY structure
/////////////////////////////////////////////////////////////////////////
typedef struct _DDS_COLORKEY
{
	unsigned int lowVal;
	unsigned int highVal;
} DDS_COLORKEY;

/////////////////////////////////////////////////////////////////////////
// DDS_SURFACE_DESC structure
/////////////////////////////////////////////////////////////////////////
typedef struct _DDS_SURFACE_DESC
{
	unsigned int	size;
	unsigned int	flags;
	unsigned int	height;
	unsigned int	width;
	unsigned int	pitch;
	unsigned int	depth;
	unsigned int	mipMapLevels;
	unsigned int	alphaBitDepth;
	unsigned int	reserved;
	unsigned int	surface;

	DDS_COLORKEY	ckDestOverlay;
	DDS_COLORKEY	ckDestBlt;
	DDS_COLORKEY	ckSrcOverlay;
	DDS_COLORKEY	ckSrcBlt;

	DDS_PIXELFORMAT format;
	DDS_CAPS		caps;

	unsigned int	textureStage;
} DDS_SURFACE_DESC;

////////////////////////////////////////////////////////////////////////
// TextureLoaderDDS class
////////////////////////////////////////////////////////////////////////
class TextureLoaderDDS : public TextureLoaderBase
{
	//------------------------------------------------------------------
	// list of friend classes
	//------------------------------------------------------------------
	/* NOTHING */

public:
	//------------------------------------------------------------------
	// public variables
	//------------------------------------------------------------------
	/* NOTHING */

	//------------------------------------------------------------------
	// public methods
	//------------------------------------------------------------------
	TextureLoaderDDS();
	virtual ~TextureLoaderDDS();
	bool Load( const char* pFilename );
	bool CheckMagic( const char* pFilename );
	bool Parse();

protected:
	//------------------------------------------------------------------
	// protected variables
	//------------------------------------------------------------------
	DDS_SURFACE_DESC mDesc;

	//------------------------------------------------------------------
	// protected methods
	//------------------------------------------------------------------
	bool IsBitMask( long r, long g, long b, long a );
	bool Parse64Bit();
    bool Parse32Bit();
    bool Parse24Bit();
	bool Parse16Bit();
	bool Parse8Bit();
    bool ParseDXT();

private:
	//------------------------------------------------------------------
	// private variables
	//------------------------------------------------------------------
	/* NOTHING */

	//------------------------------------------------------------------
	// private methods
	//------------------------------------------------------------------
	/* NOTHING */
};

} // namespace Graphics
} // namespace Asura
} // namespace Schaft

#endif // __ASURA_TEXTURE_LOADER_DDS_H__