//----------------------------------------------------------------------
// File : AsDemoApp.cpp
// Desc : Asura Demo Application
// Copyright(c) Project Asura All right reserved.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// History
//----------------------------------------------------------------------
// <---> 2010/12/04 [Pocol] VK쐬.
// <End>

//----------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------
#include "AsDemoApp.h"
#include <string>
#include <GL/glew.h>
#include <GL/glut.h>
#ifdef _WIN32
#include <GL/wglew.h>
#endif

//----------------------------------------------------------------------
// Using Statements
//----------------------------------------------------------------------
using namespace std;
using namespace Schaft::Asura::SampleProgram;


//----------------------------------------------------------------------
// Linker
//----------------------------------------------------------------------
#pragma comment( lib, "glew32.lib" )
#ifdef _WIN32
#pragma comment( lib, "glew32s.lib" )
#endif


//----------------------------------------------------------------------
// Forward Declarations
//----------------------------------------------------------------------
class DemoAppImpl;
static DemoAppImpl* GetConvertedInstance();
static void DemoAppDisplayCb();
static void DemoAppIdleCb();
static void DemoAppResizeCb( int width, int height );
static void DemoAppKeyboardDownCb( unsigned char key, int x, int y );
static void DemoAppKeyboardUpCb( unsigned char key, int x, int y );
static void DemoAppSpecialDownCb( int key, int x, int y );
static void DemoAppSpecialUpCb( int key, int x, int y );
static void DemoAppPassiveMotionCb( int x, int y );
static void DemoAppMouseCb( int button, int state, int x, int y );
static void DemoAppExit( int code );


//----------------------------------------------------------------------
// Global Variables
//----------------------------------------------------------------------
const int NUM_KEYS = 0x80;
const int NUM_MOUSE_BUTTONS = 3;


/////////////////////////////////////////////////////////////////////////
// Functions
/////////////////////////////////////////////////////////////////////////

///----------------------------------------------------------------------
///<summary>
///GLG[邩ǂ肷
///<summary>
///<returns>GLG[ꍇtrueԋp</returns>
///----------------------------------------------------------------------
bool IsGLError()
{
	bool result = false;

	GLenum error = glGetError();
	if ( error != GL_NO_ERROR )
	{
		result = true;
		switch( error )
		{
			case GL_INVALID_ENUM:
				ELOG( "GLError : Invalid Enum." );
				break;

			case GL_INVALID_VALUE:
				ELOG( "GLError : Invalid Value." );
				break;

			case GL_INVALID_OPERATION:
				ELOG( "GLError : Invalid Operation." );
				break;

			case GL_STACK_OVERFLOW:
				ELOG( "GLError : Stack Overflow." );
				break;

			case GL_STACK_UNDERFLOW:
				ELOG( "GLError : Stack Underflow." );
				break;

			case GL_OUT_OF_MEMORY:
				ELOG( "GLError : Out Of Memory." );
				break;

			default:
				ELOG( "GLError : error code=%d", error );
				break;
		}
	}
	return result;	
}


/////////////////////////////////////////////////////////////////////////
// DemoAppImpl class
/////////////////////////////////////////////////////////////////////////
class DemoAppImpl : public DemoApp
{
	//-------------------------------------------------------------------
	// list of friend classes or methods
	//-------------------------------------------------------------------
	friend void DemoAppResizeCb( int width, int height );
	friend void DemoAppKeyboardDownCb( unsigned char key, int x, int y );
	friend void DemoAppKeyboardUpCb( unsigned char key, int x, int y );
	friend void DemoAppSpecialDownCb( int key, int x, int y );
	friend void DemoAppSpecialUpCb( int key, int x, int y );
	friend void DemoAppMouseCb( int button, int state, int x, int y );
	friend void DemoAppPassiveMotionCb( int x, int y );

public:
	//-------------------------------------------------------------------
	// public variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------
	DemoAppImpl();
	~DemoAppImpl();
	int		Register( IRenderer* pRenderer );
	void	Run();
	int     GetWidth();
	int     GetHeight();
	float   GetAspectRatio();
	bool    IsKeyDown( int keycode );
	bool    IsKeyUp( int keycode );
	bool    IsMouseDown( int button );
	bool    IsMouseUp( int buttton );
	int     GetCursorX();
	int     GetCursorY();
	bool	Initialize();
	void    SetTitle( const char *title );
	void    SetSize( int width, int height );
	void	PreRender();
	void	Render();
	void    RenderAlpha();
	void	PostRender();
	void	Release();

protected:
	//-------------------------------------------------------------------
	// protected variables
	//-------------------------------------------------------------------
	std::vector< IRenderer* >	mRenderers;
	int							mWidth;
	int							mHeight;
	int                         mCursorX;
	int                         mCursorY;
	bool*                       mKeyDownStates;
	bool*                       mKeyUpStates;
	bool*                       mMouseDownStates;
	bool*                       mMouseUpStates;
	std::string                 mTitle;
	bool                        mIsCreateWindow;

	//-------------------------------------------------------------------
	// protected methods
	//-------------------------------------------------------------------
	/* NOTHING */

private:
	//-------------------------------------------------------------------
	// private variables
	//-------------------------------------------------------------------
	/* NOTHING */

	//-------------------------------------------------------------------
	// private methods
	//-------------------------------------------------------------------
	/* NOTHING */
};


///----------------------------------------------------------------------
///<summary>
///RXgN^
///</summary>
///----------------------------------------------------------------------
DemoAppImpl::DemoAppImpl()
{
	mRenderers.clear();
	mWidth = 600;
	mHeight = 480;
	mCursorX = 0;
	mCursorY = 0;
	mTitle.clear();
	mIsCreateWindow = false;
	
	if ( NUM_KEYS > 0 )
	{
		//@L[̏Ԃi[ϐ̃m
		try 
		{
			mKeyDownStates = new bool [NUM_KEYS];
			mKeyUpStates = new bool [NUM_KEYS];
		}
		catch( std::bad_alloc ba )
		{
			SAFE_DELETE_ARRAY( mKeyDownStates );
			SAFE_DELETE_ARRAY( mKeyUpStates );
			ELOG( "Error : Memory Allocate Failed. reason=%s", ba.what() );
			assert( false );
		}
		memset( mKeyDownStates, 0, sizeof(bool) * NUM_KEYS );
		memset( mKeyUpStates, 0, sizeof(bool) * NUM_KEYS );
	}
	else
	{
		mKeyDownStates =  null;
		mKeyUpStates = null;
	}

	
	if ( NUM_MOUSE_BUTTONS > 0 )
	{
		//@}EX{^̏Ԃi[ϐ̃m
		try 
		{
			mMouseDownStates = new bool [NUM_MOUSE_BUTTONS];
			mMouseUpStates = new bool [NUM_MOUSE_BUTTONS];
		}
		catch( std::bad_alloc ba )
		{
			SAFE_DELETE_ARRAY( mMouseDownStates );
			SAFE_DELETE_ARRAY( mMouseUpStates );
			ELOG( "Error : Memory Allocate Failed. reason=%s", ba.what() );
			assert( false );
		}
		memset( mMouseDownStates, 0, sizeof(bool) * NUM_MOUSE_BUTTONS );
		memset( mMouseUpStates, 0, sizeof(bool) * NUM_MOUSE_BUTTONS );
	}
	else
	{
		mMouseDownStates = null;
		mMouseUpStates = null;
	}

}

///----------------------------------------------------------------------
///<summary>
///fXgN^
///</summary>
///----------------------------------------------------------------------
DemoAppImpl::~DemoAppImpl()
{
	Release();
}

///----------------------------------------------------------------------
///<summary>
///AvP[V̏s
///</summary>
///----------------------------------------------------------------------
bool DemoAppImpl::Initialize()
{
	if ( glewInit() != GLEW_OK )
	{
		ELOG( "GLError : GLEW Initialized Failed." );
		return false;
	}

#ifdef _WIN32
	//@OFF
	wglSwapIntervalEXT( 0 ); 
#endif

	glClearColor( 100.0/255.0f, 149.0f/255.0f, 237.0f/255.0f, 1.0f );
	glShadeModel( GL_SMOOTH );
	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LEQUAL );

	for( size_t i=0; i<mRenderers.size(); ++i )
	{
		if ( !mRenderers[i]->Initialize() )
			return false;

		assert( !IsGLError() );
	}

	return true;
}

///----------------------------------------------------------------------
///<summary>
///_OȌ
///</summary>
///----------------------------------------------------------------------
void DemoAppImpl::PreRender()
{
	for( size_t i=0; i<mRenderers.size(); ++i )
	{
		mRenderers[i]->PreRender();
		assert( !IsGLError() );
	}
}

///----------------------------------------------------------------------
///<summary>
///_O
///</summary>
///----------------------------------------------------------------------
void DemoAppImpl::Render()
{
	for( size_t i=0; i<mRenderers.size(); ++i )
	{
		mRenderers[i]->Render();
		assert( !IsGLError() );
	}	
}

///----------------------------------------------------------------------
///<summary>
///ߏKvƂ̂̃_O
///</summary>
///----------------------------------------------------------------------
void DemoAppImpl::RenderAlpha()
{
	for( size_t i=0; i<mRenderers.size(); ++i )
	{
		mRenderers[i]->RenderAlpha();
		assert( !IsGLError() );
	}
}

///----------------------------------------------------------------------
///<summary>
///_Ȍ
///</summary>
///----------------------------------------------------------------------
void DemoAppImpl::PostRender()
{
	for( size_t i=0; i<mRenderers.size(); ++i )
	{
		mRenderers[i]->PostRender();
		assert( !IsGLError() );
	}	
}

///----------------------------------------------------------------------
///<summary>
///
///</summary>
///----------------------------------------------------------------------
void DemoAppImpl::Release()
{
	for( size_t i=0; i<mRenderers.size(); ++i )
	{
		SAFE_RELEASE( mRenderers[i] );
	}

	SAFE_DELETE_ARRAY( mKeyDownStates );
	SAFE_DELETE_ARRAY( mKeyUpStates );
	SAFE_DELETE_ARRAY( mMouseDownStates );
	SAFE_DELETE_ARRAY( mMouseUpStates );
	mRenderers.clear();
}

///----------------------------------------------------------------------
///<summary>
///_[o^.
///o^ꍇRun()֐̎sOɌĂяo.
///</summary>
///<param name="pRenderer">o^郌_[ւ̃|C^</param>
///<returns>o^ԍԋp</returns>
///----------------------------------------------------------------------
int DemoAppImpl::Register( IRenderer* pRenderer )
{
	mRenderers.push_back( pRenderer );
	return static_cast< int >( mRenderers.size() - 1 );
}

///----------------------------------------------------------------------
///<summary>
///AvP[Vs
///</summary>
///----------------------------------------------------------------------
void DemoAppImpl::Run()
{
	if ( mTitle.empty() )
	{
		mTitle = "Project Asura Demo Program";
	}

	//@GLUT̐ݒ
	glutInitWindowSize( mWidth, mHeight );
	glutInitDisplayMode( GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE );
	__glutCreateWindowWithExit( mTitle.c_str(), DemoAppExit );
	glutDisplayFunc( DemoAppDisplayCb );
	glutReshapeFunc( DemoAppResizeCb );
	glutIdleFunc( DemoAppIdleCb );
	glutMouseFunc( DemoAppMouseCb );
	glutPassiveMotionFunc( DemoAppPassiveMotionCb );
	glutKeyboardFunc( DemoAppKeyboardDownCb );
	glutKeyboardUpFunc( DemoAppKeyboardUpCb );
	glutSpecialFunc( DemoAppSpecialDownCb );
	glutSpecialUpFunc( DemoAppSpecialUpCb );

	//@
	if ( !Initialize() )
	{
		ELOG( "Error : Initialize Failed" );
		return;
	}

	//@C[v
	glutMainLoop();

}

///----------------------------------------------------------------------
///<summary>
///EBhE̕擾
///</summary>
///<returns>EBhE̕ԋp</returns>
///----------------------------------------------------------------------
int DemoAppImpl::GetWidth()
{
	return mWidth;
}

///----------------------------------------------------------------------
///<summary>
///EBhE̍擾
///</summary>
///<returns>EBhE̕ԋp</returns>
///----------------------------------------------------------------------
int DemoAppImpl::GetHeight()
{
	return mHeight;
}

///----------------------------------------------------------------------
///<summary>
///EBhẼAXyNg擾
///</summary>
///<returns>EBhẼAXyNgԋp</returns>
///----------------------------------------------------------------------
float DemoAppImpl::GetAspectRatio()
{
	return (float)mWidth/(float)mHeight;
}

///----------------------------------------------------------------------
///<summary>
///J[\XW擾
///</summary>
///<returns>J[\XWԋp</return>
///----------------------------------------------------------------------
int DemoAppImpl::GetCursorX()
{
	return mCursorX;
}

///----------------------------------------------------------------------
///<summary>
///J[\YW擾
///</summary>
///<returns>J[\YWԋp</returns>
///----------------------------------------------------------------------
int DemoAppImpl::GetCursorY()
{
	return mCursorY;
}

///----------------------------------------------------------------------
///<summary>
///^Cgݒ肷
///</summary>
///<param name="title">ݒ肷^Cg</param>
///----------------------------------------------------------------------
void DemoAppImpl::SetTitle( const char* title )
{
	mTitle = std::string(title);
	if ( mIsCreateWindow )
	{
		glutSetWindowTitle( mTitle.c_str() );
	}
}

///----------------------------------------------------------------------
///<summary>
///TCYݒ肷
///</summary>
///<param name="width">EBhE̕</param>
///<param name="height">EBhE̍</param>
///----------------------------------------------------------------------
void DemoAppImpl::SetSize( int width, int height )
{
	mWidth = ( width > 1 ) ? width : 1;
	mHeight = ( height > 1 ) ? height : 1;
	if ( mIsCreateWindow )
	{
		glutReshapeWindow( mWidth, mHeight );
	}
}

///----------------------------------------------------------------------
///<summary>
///L[Ă邩ǂ肷
///</summary>
///<param name="keycode">肷L[̃L[R[h</param>
///<returns>Ătrueԋp</returns>
///----------------------------------------------------------------------
bool DemoAppImpl::IsKeyDown( int keycode )
{
	int index = 0;
	if ( mKeyDownStates )
	{
		return mKeyDownStates[index];
	}
	return false;
}

///----------------------------------------------------------------------
///<summary>
///L[ꂽǂ肷
///</summary>
///<param name="keycode">肷L[̃L[R[h</param>
///<returns>ꂽtrueԋp</returns>
///----------------------------------------------------------------------
bool DemoAppImpl::IsKeyUp( int keycode )
{
	int index = 0;
	if ( mKeyUpStates )
	{
		return mKeyUpStates[index];
	}
	return false;
}

///----------------------------------------------------------------------
///<summary>
///}EX̃{^ꂽǂ肷
///</summary>
///<paranm name="button">肷{^</param>
///<returns>Ătrueԋp</returns>
///----------------------------------------------------------------------
bool DemoAppImpl::IsMouseDown( int button )
{
	int index = 0;
	if ( mMouseDownStates )
	{	
		switch( button )
		{
			case GLUT_LEFT_BUTTON:  { index = 0; } break;
			case GLUT_RIGHT_BUTTON: { index = 1; } break;
			case GLUT_MIDDLE_BUTTON:{ index = 2; } break;
			default: break;
		}
		return mMouseDownStates[index];
	}
	return false;
}

///----------------------------------------------------------------------
///<summary>
///}EX̃{^ꂽǂ肷
///</summary>
///----------------------------------------------------------------------
bool DemoAppImpl::IsMouseUp( int button )
{
	int index = 0;
	if ( mMouseUpStates )
	{
		switch( button )
		{
			case GLUT_LEFT_BUTTON:  { index = 0; } break;
			case GLUT_RIGHT_BUTTON: { index = 1; } break;
			case GLUT_MIDDLE_BUTTON:{ index = 2; } break;
			default: break;
		}
		return mMouseUpStates[index];
	}
	return false;
}


/////////////////////////////////////////////////////////////////////////
// DemoApp class
/////////////////////////////////////////////////////////////////////////

///----------------------------------------------------------------------
///<summary>
///B̃CX^X擾
///</summary>
///</returns>B̃CX^X̃|C^</returns>
///----------------------------------------------------------------------
DemoApp* DemoApp::GetInstance()
{
	static DemoAppImpl instance;
	return &instance;
}


/////////////////////////////////////////////////////////////////////////
// Internal Function for this module
/////////////////////////////////////////////////////////////////////////

///----------------------------------------------------------------------
///<summary>
///hNXփ_ELXgʂԋp.
///</summary>
///<returns>LXgɎsꍇnullԋp</returns>
///----------------------------------------------------------------------
DemoAppImpl* GetConvertedInstance()
{
	DemoAppImpl* pInstance = null;
	try
	{
		pInstance = dynamic_cast< DemoAppImpl* >( DemoApp::GetInstance() );
	}
	catch( std::bad_cast bc )
	{
		ELOG( "Error : Down Cast Failed. reason=%s", bc.what() );
		pInstance = null;
	}
	return pInstance;
}

///----------------------------------------------------------------------
///<summary>
///`揈sR[obN֐
///</summary>
///----------------------------------------------------------------------
void DemoAppDisplayCb()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	DemoAppImpl* app = GetConvertedInstance();
	if ( app )
	{
		app->PreRender();
		app->Render();
		app->RenderAlpha();
		app->PostRender();
	}

	glutSwapBuffers();
}

///----------------------------------------------------------------------
///<summary>
///TCYύXsR[obN֐
///</summary>
///<param name="width">TCYύX̃EBhE̕</param>
///<param name="height>TCYύX̃EBhE̍</param>
///----------------------------------------------------------------------
void DemoAppResizeCb( int width, int height )
{
	DemoAppImpl* app = GetConvertedInstance();
	if ( app )
	{
		app->mWidth = ( width > 1 ) ? width : 1;
		app->mHeight = ( height > 1 ) ? height : 1;
	}
}

///----------------------------------------------------------------------
///<summary>
///AChȌsR[obN֐
///</summary>
///----------------------------------------------------------------------
void DemoAppIdleCb()
{
	glutPostRedisplay();
}

///----------------------------------------------------------------------
///<summary>
///}EX{^ꂽ̏sR[obN֐
///</summary>
///<param name="button">}EX̃{^</param>
///<param name="state">{^̏</param>
///<param name="x">J[\XW</param>
///<param name="y">J[\YW</param>
///----------------------------------------------------------------------
void DemoAppMouseCb( int button, int state, int x, int y )
{
	if ( NUM_MOUSE_BUTTONS > 0 )
	{
		DemoAppImpl* app = GetConvertedInstance();
		if ( app )
		{
			app->mCursorX = x;
			app->mCursorY = y;

			//@S{^
			memset( app->mMouseDownStates, 0, sizeof(bool) * NUM_MOUSE_BUTTONS );
			memset( app->mMouseUpStates, 0, sizeof(bool) * NUM_MOUSE_BUTTONS );

			//@CfbNXZo
			int index = 0;
			switch( button )
			{
				case GLUT_LEFT_BUTTON:  { index = 0; } break;
				case GLUT_RIGHT_BUTTON: { index = 1; } break;
				case GLUT_MIDDLE_BUTTON:{ index = 2; } break;
				default: break;
			}

			//@Ԃi[
			if( state == GLUT_DOWN )
			{
				app->mMouseDownStates[index] = true;
			}
			else if( state == GLUT_UP )
			{
				app->mMouseUpStates[index] = true;
			}
		}
	}
}

///----------------------------------------------------------------------
///<summary>
///}EXړ̏sR[obN֐
///</summary>
///<param name="x">J[\XW</param>
///<param name="y">J[\YW</param>
///----------------------------------------------------------------------
void DemoAppPassiveMotionCb( int x, int y )
{
	DemoAppImpl* app = GetConvertedInstance();
	if ( app )
	{
		//@J[\̍Wi[
		app->mCursorX = x;
		app->mCursorY = y;
	}
}

///----------------------------------------------------------------------
///<summary>
///L[̏sR[obN֐
///</summary>
///<param name="key">ꂽL[</param>
///<param name="x">J[\XW</param>
///<param name="y">J[\YW</param>
///----------------------------------------------------------------------
void DemoAppKeyboardDownCb( unsigned char key, int x, int y )
{
	if( NUM_KEYS > 0 )
	{
		DemoAppImpl* app = GetConvertedInstance();
		if ( app )
		{
			//@J[\̍Wi[
			app->mCursorX = x;
			app->mCursorY = y;

			//@L[̏Ԃi[
			memset( app->mKeyDownStates, 0, sizeof(bool) * NUM_KEYS );
			if ( 0x00 <= key && key < NUM_KEYS )
			{
				app->mKeyDownStates[key] = true;
			}
		}
	}
}

///----------------------------------------------------------------------
///<summary>
///L[ꂽ̏sR[obN֐
///</summary>
///<param name="key">ꂽL[</param>
///<param name="x">J[\XW</param>
///<param name="y">J[\YW</param>
///----------------------------------------------------------------------
void DemoAppKeyboardUpCb( unsigned char key, int x, int y )
{
	if( NUM_KEYS > 0 )
	{
		DemoAppImpl* app = GetConvertedInstance();
		if ( app )
		{
			//@J[\̍Wi[
			app->mCursorX = x;
			app->mCursorY = y;

			//@L[̏Ԃi[
			memset( app->mKeyUpStates, 0, sizeof(bool) * NUM_KEYS );
			if ( 0x00 <= key && key < NUM_KEYS )
			{
				app->mKeyUpStates[key] = false;
			}
		}
	}
}

///----------------------------------------------------------------------
///<summary>
///L[ꂽ̏sR[obN֐
///</summary>
///<param name="key">ꂽL[</param>
///<param name="x">J[\XW</param>
///<param name="y">J[\YW</param>
///----------------------------------------------------------------------
void DemoAppSpecialDownCb( int key, int x, int y )
{
	if( NUM_KEYS > 0x80 )
	{
		DemoAppImpl* app = GetConvertedInstance();
		if ( app )
		{
			//@J[\̍Wi[
			app->mCursorX = x;
			app->mCursorY = y;

			/* DO NOTHING */
		}
	}
}

///----------------------------------------------------------------------
///<summary>
///L[ꂽ̏sR[obN֐
///</summary>
///<param name="key">ꂽL[</param>
///<param name="x">J[\XW</param>
///<param name="y">J[\YW</param>
///----------------------------------------------------------------------
void DemoAppSpecialUpCb( int key, int x, int y )
{
	if( NUM_KEYS > 0 )
	{
		DemoAppImpl* app = GetConvertedInstance();
		if ( app )
		{
			//@J[\̍Wi[
			app->mCursorX = x;
			app->mCursorY = y;

			/* DO NOTHING */
		}
	}
}

///----------------------------------------------------------------------
///<summary>
///AvP[VȈ
///</summary>
///----------------------------------------------------------------------
void DemoAppExit( int code )
{
	DemoAppImpl* app = GetConvertedInstance();
	SAFE_RELEASE( app );

	std::exit( code );
}

// End Of File.