//----------------------------------------------------------------------
// File : SampleRenderer.cpp
// Desc : Sample Renderer
// Copyright(c) Project Asura All right reserved.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// History
//----------------------------------------------------------------------
// <---> 2011/02/19 [Pocol] VK쐬.
// <End>

//----------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------
#include "SampleRenderer.h"
#include <GL/glew.h>
#include <GL/glut.h>
#include "../SHCalculator.h"

#pragma comment( lib, "glew32.lib" )

//----------------------------------------------------------------------
// Using Statements
//----------------------------------------------------------------------
using namespace Schaft::Asura;
using namespace Schaft::Asura::Graphics;
using namespace Schaft::Asura::Math;
using namespace Schaft::Asura::SampleProgram;

namespace Schaft {
namespace Asura {
namespace SampleProgram {

///----------------------------------------------------------------------
///<summary>
///ftHgl̐ݒ
///</summary>
///----------------------------------------------------------------------
SampleRenderer::SampleRenderer()
{
    mWorld.Identity();
    mView.Identity();
    mProjection.Identity();

    mLightPosition = Vector3( 100.0f, 100.0f, 100.0f );
}

///----------------------------------------------------------------------
///<summary>
///fXgN^
///</summary>
///----------------------------------------------------------------------
SampleRenderer::~SampleRenderer()
{
    /* DO_NOTHING */
}

///----------------------------------------------------------------------
///<summary>
///
///</summary>
///<returns>ɐtrueԋp</returns>
///----------------------------------------------------------------------
bool SampleRenderer::Initialize()
{
    bool result = true;

    mCameraPosition = Vector3( 2.5f, 5.0f, -5.0f );
    Vector3 dir = Vector3( 0.0f, 0.0f, 0.0f );
    Vector3 up = Vector3( 0.0f, 1.0f, 0.0f );

    mWorld.Identity();
    mView = CreateLookAt( mCameraPosition, dir, up );
    mProjection = CreatePerspectiveFieldOfView( Math::PiOver4, DemoApp::GetInstance()->GetAspectRatio(), 0.1f, 1000.0f );

    CreateObject( 5, -1 );

    return result;
}

///----------------------------------------------------------------------
///<summary>
///
///</summary>
///----------------------------------------------------------------------
void SampleRenderer::Release()
{
    for( int i=0; i<3; i++ )
    {
        if ( mVB[i] )
        {
            glDeleteBuffers( 1, &mVB[i] );
            mVB[i] = 0;
        }
    }

    if ( mIB )
    {
        glDeleteBuffers( 1, &mIB );
        mIB = 0;
    }
}

///----------------------------------------------------------------------
///<summary>
///_OȌ
///</summary>
///----------------------------------------------------------------------
void SampleRenderer::PreRender()
{
    /* DO_NOTHING */
}

///----------------------------------------------------------------------
///<summary>
///ŝ̃_O
///</summary>
///----------------------------------------------------------------------
void SampleRenderer::Render()
{
    // r[|[g̐ݒ
    glViewport( 0, 0, DemoApp::GetInstance()->GetWidth(), DemoApp::GetInstance()->GetHeight() );

    // fr[svZ
    Matrix worldView = mWorld * mView;

    glEnable( GL_COLOR_MATERIAL );
    glColorMaterial( GL_FRONT_AND_BACK, GL_DIFFUSE );

    // fr[sݒ
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glLoadMatrixf( worldView );

    // ˉesݒ
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glLoadMatrixf( mProjection );

    // obt@IuWFNg̐ݒ
    // _obt@
    glEnableClientState( GL_VERTEX_ARRAY );
    glBindBuffer( GL_ARRAY_BUFFER, mVB[0] );
    glVertexPointer( 3, GL_FLOAT, 0, null );

    // @obt@
    glEnableClientState( GL_NORMAL_ARRAY );
    glBindBuffer( GL_ARRAY_BUFFER, mVB[1] );
    glNormalPointer( GL_FLOAT, 0, null );

    // J[obt@
    glEnableClientState( GL_COLOR_ARRAY );
    glBindBuffer( GL_ARRAY_BUFFER, mVB[2] );
    glColorPointer( 4, GL_FLOAT, 0, null );

    // CfbNXobt@
    glEnableClientState( GL_INDEX_ARRAY );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, mIB );
    glIndexPointer( GL_INT, 0, null );

    // `
    glDrawElements( GL_TRIANGLES, mNumIndices, GL_UNSIGNED_INT, NULL );

    // Еt
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
    glBindBuffer( GL_ARRAY_BUFFER, 0 );
    glDisableClientState( GL_INDEX_ARRAY );
    glDisableClientState( GL_COLOR_ARRAY );
    glDisableClientState( GL_NORMAL_ARRAY );
    glDisableClientState( GL_VERTEX_ARRAY );

}

///----------------------------------------------------------------------
///<summary>
///̂̃_O
///</summary>
///----------------------------------------------------------------------
void SampleRenderer::RenderAlpha()
{
    /* DO_NOTHING */
}

///----------------------------------------------------------------------
///<summary>
///_Ȍ
///</summary>
///----------------------------------------------------------------------
void SampleRenderer::PostRender()
{
    /* DO_NOTHING */
}

///----------------------------------------------------------------------
///<summary>
///_OpIuWFNg𐶐
///</summary>
///----------------------------------------------------------------------
void SampleRenderer::CreateObject( int l, int m )
{
    const int x = 256;
    const int y = 256;

    int numVertices = ((x+1) * (y+1));
    int numFaces = 2 * x * y;
    mNumIndices = 3 * numFaces;

    std::vector< Vector3 > vertices;
    std::vector< Vector3 > normals;
    std::vector< Vector4 > colors;

    //m
    vertices.resize( numVertices );
    normals.resize( numVertices );
    colors.resize( numVertices );
    
    int count = 0;
    for( int j=0; j<=y; j++ )
    {
        for( int i=0; i<=x; i++ )
        {
            float phi = (float)i * 2.0f * Math::Pi / ( x - Math::Pi );
            float theta = (float)j * Math::Pi / y;

            // ʒa֐̒lvZ
            float sh = SHCalculator::Compute( l, m, theta, phi );
            
            // ̒_WZo
            vertices[count].x = sinf( theta ) * cosf( phi );
            vertices[count].y = cosf( theta );
            vertices[count].z = sinf( theta ) * sinf( phi );

            // @xNgZo
            normals[count].x = sinf( theta ) * cosf( phi );
            normals[count].y = cosf( theta );
            normals[count].z = sinf( theta ) * sinf( phi );

            // ̗̈͐ԁC̗̈͗΂ŐFÂ
            if ( sh < 0 )
            {
                colors[count].x = 1.0f * abs( sh );
                colors[count].y = 0.0f * abs( sh ); 
                colors[count].z = 0.0f * abs( sh );
            }
            else
            {
                colors[count].x = 0.0f * abs( sh );
                colors[count].y = 1.0f * abs( sh );
                colors[count].z = 0.0f * abs( sh );
            }
            colors[count].w = 1.0f;

            count++;
        }
    }

    std::vector< uint32_t > indices;
    indices.resize( mNumIndices );
    count = 0;

    // _CfbNX߂
    for( uint32_t j=0; j<y; j++ )
    {
        for( uint32_t i=0; i<x; i++ )
        {
            indices[count] = ( j + 0 ) * ( x + 1 ) + ( i + 0 ); count++;
            indices[count] = ( j + 0 ) * ( x + 1 ) + ( i + 1 ); count++;
            indices[count] = ( j + 1 ) * ( x + 1 ) + ( i + 0 ); count++;
            indices[count] = ( j + 0 ) * ( x + 1 ) + ( i + 1 ); count++;
            indices[count] = ( j + 1 ) * ( x + 1 ) + ( i + 1 ); count++;
            indices[count] = ( j + 1 ) * ( x + 1 ) + ( i + 0 ); count++;
        }
    }

    // obt@IuWFNg
    // _obt@
    glGenBuffers( 1, &mVB[0] );
    glBindBuffer( GL_ARRAY_BUFFER, mVB[0] );
    glBufferData( GL_ARRAY_BUFFER, sizeof( Vector3 ) * numVertices, &vertices[0], GL_STATIC_DRAW );

    // @obt@
    glGenBuffers( 1, &mVB[1] );
    glBindBuffer( GL_ARRAY_BUFFER, mVB[1] );
    glBufferData( GL_ARRAY_BUFFER, sizeof( Vector3 ) * numVertices, &normals[0], GL_STATIC_DRAW );

    // J[obt@
    glGenBuffers( 1, &mVB[2] );
    glBindBuffer( GL_ARRAY_BUFFER, mVB[2] );
    glBufferData( GL_ARRAY_BUFFER, sizeof( Vector4 ) * numVertices, &colors[0], GL_STATIC_DRAW );

    // CfbNXobt@
    glGenBuffers( 1, &mIB );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, mIB );
    glBufferData( GL_ELEMENT_ARRAY_BUFFER, sizeof( uint32_t ) * mNumIndices, &indices[0], GL_STATIC_DRAW );

    // 
    vertices.clear();
    normals.clear();
    colors.clear();
    indices.clear();
}

} // namespace SampleProgram
} // namespace Asura
} // namespace Schaft