//-----------------------------------------------------------------------
// File : SHCalculator.cpp
// Desc : Shperical Harmonics Calculator
// Copyright(c) Project Asura All right reserved.
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// History
//-----------------------------------------------------------------------
// <---> 2011/02/19 [Pocol] VK쐬.
// <End>

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include "SHCalculator.h"

//-----------------------------------------------------------------------
// Using Statements
//-----------------------------------------------------------------------
using namespace Schaft::Asura;
using namespace Schaft::Asura::Math;

namespace Schaft {
namespace Asura {
namespace SampleProgram {

/////////////////////////////////////////////////////////////////////////
// SHCalculator class
/////////////////////////////////////////////////////////////////////////

///----------------------------------------------------------------------
///<summary>
///ʒa֐vZ
///</summary>
///<param name="l">̐w肷</param>
///<param name="m">-ll܂ł̐w肷</param>
///<param name="theta">px</param>
///<param name="phi">px</param>
///<returns>SH֐vZʂԋp</returns>
///----------------------------------------------------------------------
float SHCalculator::Compute( int l, int m, float theta, float phi )
{
    assert( l >= 0 );
    assert( -l <= m && m <= l );

    float result = ScalingFactor( l, m );

    if ( 0 < m )
    { result *= sqrtf( 2.0f ) * cosf( m * phi ) * LegendrePolynomials( l, m, cosf( theta ) ); }
    else if ( m < 0 )
    { result *= sqrtf( 2.0f ) * sinf( -m * phi ) * LegendrePolynomials( l, -m, cosf( theta ) ); }
    else
    { result *= LegendrePolynomials( l, m, cosf( theta ) ); }

    return result;
}

///----------------------------------------------------------------------
///<summary>
///K邽߂̃XP[OW߂
///</summary>
///<param name="l">̐w</param>
///<param name="m">-ll܂ł̐w</param>
///<returns>vZlԋp</returns>
///----------------------------------------------------------------------
float SHCalculator::ScalingFactor( int l, int m )
{
    float lpm = 1.0f;
    float lnm = 1.0f;

    if ( m < 0 ) { m = -m; }

    for( int i=(l-m); 0<i; i-- ) { lnm *= i; }
    for( int i=(l+m); 0<i; i-- ) { lpm *= i; }

    return sqrtf( ( ( 2.0f * l + 1.0f ) * lnm ) / ( 4.0f * Math::Pi * lpm ) );
}

///----------------------------------------------------------------------
///<summary>
///Wh֐vZ
///</summary>
///<param name="l"></param>
///<param name="m"></param>
///<param name="x"></param>
///<returns>vZlԋp</returns>
///----------------------------------------------------------------------
float SHCalculator::LegendrePolynomials( int l, int m, float x )
{
    float pmm = 1.0f;

    if ( 0 < m )
    {
        float somx2 = sqrtf( ( 1.0f - x ) * ( 1.0 + x ) );
        float fact = 1.0f;
        for( int i=1; i<=m; i++ ) 
        {
            pmm *= (-fact) * somx2;
            fact += 2.0f;
        }
    }
    if ( l == m ) { return pmm; }

    float pmmp1 = x * ( 2.0f * m + 1.0f ) * pmm;
    if ( l == (m + 1) ) { return pmmp1; }

    float pll = 0.0f;

    for( int ll=(m + 2); ll<=l; ll++ )
    {
        pll = ( (2.0f * ll - 1.0f ) * x * pmmp1 - ( ll + m - 1.0f ) * pmm ) / (ll - m);
        pmm = pmmp1;
        pmmp1 = pll;
    }

    return pll;
}


} // namespace SampleProgram
} // namespace Asura
} // namespace Schaft